# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from notebook.services.kernels.kernelmanager import MappingKernelManager
from ..cell.parser import APICellParser

class SeedingMappingKernelManager(MappingKernelManager):
    @property
    def seed_kernelspec(self):
        '''
        Gets the kernel spec name required to run the seed notebook. Returns
        None if no seed notebook exists.
        '''
        if hasattr(self, '_seed_kernelspec'):
            return self._seed_kernelspec

        if self.parent.seed_notebook:
            self._seed_kernelspec = self.parent.seed_notebook['metadata']['kernelspec']['name']
        else:
            self._seed_kernelspec = None

        return self._seed_kernelspec

    @property
    def seed_source(self):
        '''
        Gets the source of the seed notebook in cell order. Returns None if no
        seed notebook exists.
        '''
        if hasattr(self, '_seed_source'):
            return self._seed_source

        if self.parent.seed_notebook:
            self._seed_source = [
                cell['source'] for cell in self.parent.seed_notebook.cells
                if cell['cell_type'] == 'code'
            ]
        else:
            self._seed_source = None

        return self._seed_source

    def start_seeded_kernel(self, *args, **kwargs):
        '''
        Start a kernel using the language specified in the seed notebook. If
        there is no seed notebook,  start a kernel using the other parameters
        specified.
        '''
        self.start_kernel(kernel_name=self.seed_kernelspec, *args, **kwargs)

    def start_kernel(self, *args, **kwargs):
        '''
        Starts a kernel and then optionally executes a list of code cells on it
        before returning its ID.
        '''
        kernel_id = super(MappingKernelManager, self).start_kernel(*args, **kwargs)

        if kernel_id and self.seed_source is not None:
            # Only run source if the kernel matches
            kernel = self.get_kernel(kernel_id)
            if kernel.kernel_name == self.seed_kernelspec:
                # Connect to the kernel and pump in the content of the notebook
                # before returning the kernel ID to the requesting client
                client = kernel.client()
                if self.parent.api == 'notebook-http':
                    client.start_channels()
                    client.wait_for_ready()
                for code in self.seed_source:
                    # Execute every code cell and wait for each to succeed or fail
                    api_cell_parser = APICellParser(self.seed_kernelspec)
                    if not api_cell_parser.is_api_cell(code) and not api_cell_parser.is_api_response_cell(code) :
                        client.execute(code)
                        msg = client.shell_channel.get_msg(block=True)
                        if msg['content']['status'] != 'ok':
                            # Shutdown the channels to remove any lingering ZMQ messages
                            client.stop_channels()
                            # Shutdown the kernel
                            self.shutdown_kernel(kernel_id)
                            raise RuntimeError('Error seeding kernel memory')
                # Shutdown the channels to remove any lingering ZMQ messages
                client.stop_channels()
        return kernel_id



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from tornado.locks import Semaphore
from tornado import gen

class KernelPool(object):
    '''
    Spawns a pool of kernels and cleans them up when shut down.
    '''
    def __init__(self, prespawn_count, kernel_manager):
        self.kernel_manager = kernel_manager
        # Make sure we've got a int
        if not prespawn_count:
            prespawn_count = 0
        for _ in range(prespawn_count):
            self.kernel_manager.start_seeded_kernel()

    def shutdown(self):
        kids = self.kernel_manager.list_kernel_ids()
        for kid in kids:
            self.kernel_manager.shutdown_kernel(kid, now=True)

class ManagedKernelPool(KernelPool):
    '''
    Spawns a pool of kernels. Manages access to individual kernels using a
    borrower/lender pattern. Cleans them all up when shut down.
    '''
    def __init__(self, prespawn_count, kernel_manager):
        # Make sure there's at least one kernel as a delegate
        if not prespawn_count:
            prespawn_count = 1

        super(ManagedKernelPool, self).__init__(prespawn_count, kernel_manager)

        self.kernel_clients = {}
        self.on_recv_funcs = {}
        self.pool_index = 0
        self.kernel_pool = []

        kernel_ids = self.kernel_manager.list_kernel_ids()
        self.kernel_semaphore = Semaphore(len(kernel_ids))

        # Connect to any prespawned kernels
        for kernel_id in kernel_ids:
            self.kernel_clients[kernel_id] = kernel_manager.get_kernel(kernel_id).client()
            self.kernel_pool.append(kernel_id)
            iopub = self.kernel_manager.connect_iopub(kernel_id)
            iopub.on_recv(self.create_on_reply(kernel_id))

    @gen.coroutine
    def acquire(self):
        '''
        Returns a kernel client and id for use and removes the kernel the resource pool.
        Kernels must be returned via the release method.
        :return: Returns a kernel client and a kernel id
        '''
        yield self.kernel_semaphore.acquire()
        kernel_id = self.kernel_pool[0]
        del self.kernel_pool[0]
        raise gen.Return((self.kernel_clients[kernel_id], kernel_id))

    def release(self, kernel_id):
        '''
        Returns a kernel back to the resource pool.
        :param kernel_id: Id of the kernel to return to the pool
        '''
        self.kernel_pool.append(kernel_id)
        self.kernel_semaphore.release()

    def _on_reply(self, kernel_id, msg_list):
        idents, msg_list = self.kernel_clients[kernel_id].session.feed_identities(msg_list)
        msg = self.kernel_clients[kernel_id].session.deserialize(msg_list)
        self.on_recv_funcs[kernel_id](msg)

    def create_on_reply(self, kernel_id):
        '''
        The lambda is used to handle a specific reply per kernel and provide a unique stack scope per invocation.
        '''
        return lambda msg_list: self._on_reply(kernel_id, msg_list)

    def on_recv(self, kernel_id, func):
        '''
        Registers a callback for io_pub messages for a particular kernel.
        This is needed to avoid having multiple callbacks per kernel client.
        :param kernel_id: Id of the kernel
        :param func: Callback function to handle the message
        '''
        self.on_recv_funcs[kernel_id] = func

    def shutdown(self):
        '''
        Shuts down all kernels in the pool and in the kernel manager.
        '''
        for kid in self.kernel_clients:
            self.kernel_clients[kid].stop_channels()
            self.kernel_manager.shutdown_kernel(kid, now=True)

        # Any remaining kernels that were not created for our pool should be shutdown
        super(ManagedKernelPool, self).shutdown()



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

LAST_MESSAGE_TO_CLIENT = 'last_message_to_client'
LAST_MESSAGE_TO_KERNEL = 'last_message_to_kernel'
LAST_TIME_STATE_CHANGED = 'last_time_state_changed'
BUSY = 'busy'
CONNECTIONS = 'connections'
LAST_CLIENT_CONNECT = 'last_client_connect'
LAST_CLIENT_DISCONNECT = 'last_client_disconnect'

default_activity_values = [(LAST_MESSAGE_TO_CLIENT, None),
    (LAST_MESSAGE_TO_KERNEL, None),
    (LAST_TIME_STATE_CHANGED , None),
    (BUSY , False),
    (CONNECTIONS , 0),
    (LAST_CLIENT_CONNECT , None),
    (LAST_CLIENT_DISCONNECT , None)
]
class ActivityManager(object):
    '''Represents a store of activity values for kernels. There is a singleton instance
    called `activity` created in this module.'''
    def __init__(self):
        self.values = {}
        # A blacklist of kernels that have been removed so no more activities cannot sneak in.
        # This can happen when a kernel is deleted, and a websocket is later disconnected
        self.ignore = set()
        self.dummy_map = {}
        self.populate_kernel_with_defaults(self.dummy_map)

    def populate_kernel_with_defaults(self, activity_values):
        '''Sets the default value for known activities being recorded.
        '''
        for value in default_activity_values:
            activity_values[value[0]] = value[1]

    def get_map_for_kernel(self, kernel_id):
        '''Gets a map for a kernel. This method will always return a map, even if the kernel has been removed.
        In the event the kernel has been removed, the activities will not be recorded.
        '''
        if kernel_id in self.ignore:
            return self.dummy_map

        if not kernel_id in self.values:
            self.values[kernel_id] =  {}
            self.populate_kernel_with_defaults(self.values[kernel_id])

        return self.values[kernel_id]

    def publish(self, kernel_id, activity_type, value=None):
        '''Sets the value stored for *activity_type*. If the activity_type is not found, it
        will be created and assigned to *value*.
        '''
        self.get_map_for_kernel(kernel_id)[activity_type] = value;


    def increment_activity(self, kernel_id, activity_type):
        '''Increments the value stored for *activity_type*. If the value currently stored
        is not an int, an TypeError will be raised. If the activity_type is not found, a
        KeyError will be raised.
        '''
        self.get_map_for_kernel(kernel_id)[activity_type] += 1;

    def decrement_activity(self, kernel_id, activity_type):
        '''Decrements the value stored for *activity_type*. If the value currently stored
        is not an int, an TypeError will be raised. If the activity_type is not found, a
        KeyError will be raised.
        '''
        self.get_map_for_kernel(kernel_id)[activity_type] -= 1;

    def remove(self, kernel_id):
        '''Removes the activities for a kernel.'''
        if kernel_id in self.values:
            del self.values[kernel_id]
            self.ignore.add(kernel_id)

    def get(self):
        '''Returns a mapping of kernel_id to activity values.
        '''
        return self.values

activity = ActivityManager()



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import json
import tornado
import notebook.services.kernels.handlers as notebook_handlers
from ...mixins import TokenAuthorizationMixin, CORSMixin, JSONErrorsMixin
from ...services.activity.manager import activity, LAST_MESSAGE_TO_CLIENT, LAST_MESSAGE_TO_KERNEL, LAST_TIME_STATE_CHANGED, BUSY, CONNECTIONS, LAST_CLIENT_CONNECT, LAST_CLIENT_DISCONNECT
from datetime import datetime

class MainKernelHandler(TokenAuthorizationMixin,
                        CORSMixin,
                        JSONErrorsMixin,
                        notebook_handlers.MainKernelHandler):
    def post(self):
        '''
        Honors the max number of allowed kernels configuration setting. Raises
        402 (for lack of a better HTTP error code) if at the limit.
        '''
        max_kernels = self.settings['kg_max_kernels']
        if max_kernels is not None:
            km = self.settings['kernel_manager']
            kernels = km.list_kernels()
            if len(kernels) >= max_kernels:
                raise tornado.web.HTTPError(402, 'Resource Limit')

        super(MainKernelHandler, self).post()

    def get(self):
        '''
        Denies returning a list of running kernels unless explicitly
        enabled, instead returning a 403 error indicating that the list is
        permanently forbidden.
        '''
        if not self.settings.get('kg_list_kernels'):
            raise tornado.web.HTTPError(403, 'Forbidden')
        else:
            super(MainKernelHandler, self).get()

    def get_json_body(self):
        '''
        Use the specified default kernel name when one is not included in the
        JSON body of the request.
        '''
        model = super(MainKernelHandler, self).get_json_body()
        if 'kg_default_kernel_name' in self.settings and self.settings['kg_default_kernel_name'] is not '':
            if model is None:
                model = {
                    'name': self.settings['kg_default_kernel_name']
                }
            else:
                model.setdefault('name', self.settings['kg_default_kernel_name'])
        return model

class KernelHandler(TokenAuthorizationMixin,
                    CORSMixin,
                    JSONErrorsMixin,
                    notebook_handlers.KernelHandler):
    '''This class will remove a kernel from the activity store when it is delted.
    '''
    def delete(self, kernel_id):
        if self.settings.get('kg_list_kernels'):
            activity.remove(kernel_id)
        super(KernelHandler, self).delete(kernel_id)

class ZMQChannelsHandler(TokenAuthorizationMixin,
                         notebook_handlers.ZMQChannelsHandler):
    '''This class listens for messages coming to and from the kernel to provide metrics
    about kernel usage.
    '''
    def open(self, kernel_id):
        if self.settings.get('kg_list_kernels'):
            activity.publish(self.kernel_id, LAST_CLIENT_CONNECT, datetime.now().isoformat())
            activity.increment_activity(self.kernel_id, CONNECTIONS)
        super(ZMQChannelsHandler, self).open(kernel_id)

    def on_close(self):
        if self.settings.get('kg_list_kernels'):
            activity.publish(self.kernel_id, LAST_CLIENT_DISCONNECT, datetime.now().isoformat())
            activity.decrement_activity(self.kernel_id, CONNECTIONS)
        super(ZMQChannelsHandler, self).on_close()

    def _on_zmq_reply(self, stream, msg_list):
        if self.settings.get('kg_list_kernels'):
            msg_metadata = json.loads(msg_list[3].decode('UTF-8'))
            # If the message coming back is a status message, we need to inspect it
            if msg_metadata['msg_type'] == 'status':
                msg_content = json.loads(msg_list[6].decode('UTF-8'))
                # If the status is busy, set the busy to True
                if msg_content['execution_state'] == 'busy':
                    activity.publish(self.kernel_id, BUSY, True)
                # Else if the status is idle, set the busy to False
                elif msg_content['execution_state'] == 'idle':
                    activity.publish(self.kernel_id, BUSY, False)
                # Record the time the state was changed
                activity.publish(self.kernel_id, LAST_TIME_STATE_CHANGED, datetime.now().isoformat())

            activity.publish(self.kernel_id, LAST_MESSAGE_TO_CLIENT, datetime.now().isoformat())
        super(ZMQChannelsHandler, self)._on_zmq_reply(stream, msg_list)

    def on_message(self, msg):
        if self.settings.get('kg_list_kernels'):
            activity.publish(self.kernel_id, LAST_MESSAGE_TO_KERNEL, datetime.now().isoformat())
        super(ZMQChannelsHandler, self).on_message(msg)


default_handlers = []
for path, cls in notebook_handlers.default_handlers:
    if cls.__name__ in globals():
        # Use the same named class from here if it exists
        default_handlers.append((path, globals()[cls.__name__]))
    else:
        # Gen a new type with CORS and token auth
        bases = (TokenAuthorizationMixin, CORSMixin, cls)
        default_handlers.append((path, type(cls.__name__, bases, {})))



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import json
try:
    # py3
    from http.client import responses
except ImportError:
    from httplib import responses

class CORSMixin(object):
    '''
    Adds CORS headers to tornado.web.RequestHandlers.
    '''
    SETTINGS_TO_HEADERS = {
        'kg_allow_credentials': 'Access-Control-Allow-Credentials',
        'kg_allow_headers': 'Access-Control-Allow-Headers',
        'kg_allow_methods': 'Access-Control-Allow-Methods',
        'kg_allow_origin': 'Access-Control-Allow-Origin',
        'kg_expose_headers' : 'Access-Control-Expose-Headers',
        'kg_max_age': 'Access-Control-Max-Age'
    }
    def set_default_headers(self):
        '''
        Sets the CORS headers as the default for all responses. Disable CSP
        for the API.
        '''
        super(CORSMixin, self).set_default_headers()

        # Add CORS headers after default if they have a non-blank value
        for settings_name, header_name in self.SETTINGS_TO_HEADERS.items():
            header_value = self.settings.get(settings_name)
            if header_value:
                self.set_header(header_name, header_value)

        # Don't set CSP: we're not serving frontend media types only JSON
        self.clear_header('Content-Security-Policy')

class TokenAuthorizationMixin(object):
    '''
    Adds token authentication to tornado.web.RequestHandlers and
    tornado.websocket.WebsocketHandlers.
    '''
    def prepare(self):
        '''
        If kg_auth_token is present in the application settings,
        raises 401 if the request Authorization header does not contain
        the value "token <kg_auth_token>".
        '''
        server_token = self.settings.get('kg_auth_token')
        if server_token:
            client_token = self.request.headers.get('Authorization')
            if client_token != 'token %s' % server_token:
                return self.send_error(401)
        return super(TokenAuthorizationMixin, self).prepare()

class JSONErrorsMixin(object):
    '''
    Outputs all errors as JSON in the same format as the
    notebook.base.handlers.json_errors decorator. Avoids having to (re-)decorate
    everything that the kernel gateway overrides. Also avoids rendering errors
    as a human readable HTML pages in cases where the decorator is not used
    in the notebook code base.
    '''
    def write_error(self, status_code, **kwargs):
        '''
        Overrides the HTML renderer in the notebook server to force all errors
        to JSON format.
        '''
        exc_info = kwargs.get('exc_info')
        message = ''
        reason = responses.get(status_code, 'Unknown HTTP Error')
        if exc_info:
            exception = exc_info[1]
            # Get the custom message, if defined
            try:
                message = exception.log_message % exception.args
            except Exception:
                pass

            # Construct the custom reason, if defined
            custom_reason = getattr(exception, 'reason', '')
            if custom_reason:
                reason = custom_reason

        self.set_header('Content-Type', 'application/json')
        self.set_status(status_code)
        reply = dict(reason=reason, message=message)
        self.finish(json.dumps(reply))



"""A base class session manager."""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import uuid

from tornado import web

from traitlets.config.configurable import LoggingConfigurable
from ipython_genutils.py3compat import unicode_type

class SessionManager(LoggingConfigurable):
    def __init__(self, kernel_manager, *args, **kwargs):
        self.kernel_manager = kernel_manager
        self._sessions = []
        self._columns = ['session_id', 'path', 'kernel_id']

    def session_exists(self, path):
        """Check to see if the session for a given notebook exists"""
        return bool([item for item in self._sessions if item['path'] == path])

    def new_session_id(self):
        """Create a uuid for a new session"""
        return unicode_type(uuid.uuid4())

    def create_session(self, path=None, kernel_name=None):
        """Creates a session and returns its model"""
        session_id = self.new_session_id()
        # allow nbm to specify kernels cwd
        kernel_id = self.kernel_manager.start_kernel(path=path,
                                                     kernel_name=kernel_name)
        return self.save_session(session_id, path=path,
                                 kernel_id=kernel_id)

    def save_session(self, session_id, path=None, kernel_id=None):
        """Saves the items for the session with the given session_id

        Given a session_id (and any other of the arguments), this method
        creates a row in the sqlite session database that holds the information
        for a session.

        Parameters
        ----------
        session_id : str
            uuid for the session; this method must be given a session_id
        path : str
            the path for the given notebook
        kernel_id : str
            a uuid for the kernel associated with this session

        Returns
        -------
        model : dict
            a dictionary of the session model
        """
        self._sessions.append({'session_id': session_id,
                               'path':path,
                               'kernel_id': kernel_id})

        return self.get_session(session_id=session_id)

    def get_session_by_key(self, key, val):
        s = [item for item in self._sessions if item[key] == val]
        return None if not s else s[0]

    def get_session(self, **kwargs):
        """Returns the model for a particular session.

        Takes a keyword argument and searches for the value in the session
        database, then returns the rest of the session's info.

        Parameters
        ----------
        **kwargs : keyword argument
            must be given one of the keywords and values from the session database
            (i.e. session_id, path, kernel_id)

        Returns
        -------
        model : dict
            returns a dictionary that includes all the information from the
            session described by the kwarg.
        """
        if not kwargs:
            raise TypeError("must specify a column to query")

        for param in kwargs.keys():
            if param not in self._columns:
                raise TypeError("No such column: %r", param)

        #multiple columns are never passed into kwargs so just using the
        #first and only one.
        column = list(kwargs.keys())[0]
        row = self.get_session_by_key(column, kwargs[column])

        if not row:
            raise web.HTTPError(404, u'Session not found: %s' % kwargs[column])

        return self.row_to_model(row)

    def update_session(self, session_id, **kwargs):
        """Updates the values in the session database.

        Changes the values of the session with the given session_id
        with the values from the keyword arguments.

        Parameters
        ----------
        session_id : str
            a uuid that identifies a session in the sqlite3 database
        **kwargs : str
            the key must correspond to a column title in session database,
            and the value replaces the current value in the session
            with session_id.
        """
        if not kwargs:
            # no changes
            return

        row = self.get_session_by_key('session_id', session_id)

        if not row:
            raise KeyError

        self._sessions.remove(row)

        if 'path' in kwargs:
            row['path'] = kwargs['path']

        if 'kernel_id' in kwargs:
            row['kernel_id'] = kwargs['kernel_id']

        self._sessions.append(row)


    def row_to_model(self, row):
        """Takes sqlite database session row and turns it into a dictionary"""
        if row['kernel_id'] not in self.kernel_manager:
            # The kernel was killed or died without deleting the session.
            # We can't use delete_session here because that tries to find
            # and shut down the kernel.
            self._sessions.remove(row)
            raise KeyError

        model = {
            'id': row['session_id'],
            'notebook': {
                'path': row['path']
            },
            'kernel': self.kernel_manager.kernel_model(row['kernel_id'])
        }
        return model

    def list_sessions(self):
        """Returns a list of dictionaries containing all the information from
        the session database"""

        l = [self.row_to_model(r) for r in self._sessions]
        return l

    def delete_session(self, session_id):
        """Deletes the row in the session database with given session_id"""
        # Check that session exists before deleting
        s = self.get_session_by_key('session_id', session_id)
        if not s:
            raise KeyError

        self.kernel_manager.shutdown_kernel(s['kernel_id'])
        self._sessions.remove(s)



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import tornado.web
import json
from .builders import *
from ...mixins import TokenAuthorizationMixin, CORSMixin, JSONErrorsMixin

class SwaggerSpecHandler(TokenAuthorizationMixin,
                         CORSMixin,
                         JSONErrorsMixin,
                         tornado.web.RequestHandler):
    '''A tornado handler to serve requests for a swagger specification.
    Given a collection of cell sources, notebook title and kernel name, this
    handler will generate a swagger specification describing the API.
    '''
    output = '{}'

    def initialize(self, title, source_cells, kernel_spec):
        spec_builder = SwaggerSpecBuilder(kernel_spec)
        for source_cell in source_cells:
            spec_builder.add_cell(source_cell)
        spec_builder.set_title(title)
        self.output = json.dumps(spec_builder.build())

    def get(self, **kwargs):
        self.set_header('Content-Type', 'application/json')
        self.set_status(200)
        self.finish(self.output)



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from ..cell.parser import APICellParser
from os import path

class SwaggerSpecBuilder(object):
    '''
    Builds a swagger specification by calling the various methods to add more
    components to the specification.
    '''
    def __init__(self, kernel_spec):
        self.cell_parser = APICellParser(kernel_spec)
        self.value = { 'swagger' : '2.0', 'paths' : {}, 'info' : {'version' : '0.0.0', 'title' : 'Default Title'} }

    def add_cell(self, cell_source):
        '''Parses the cell for API comments. If some are found then the HTTP verb
        and endpoint for the cell are added to the swagger specification.
        '''
        if self.cell_parser.is_api_cell(cell_source):
            path_name, verb = self.cell_parser.get_cell_endpoint_and_verb(cell_source)
            path_value = {
                'responses' : {
                    200 : { 'description': 'Success'}
                }
            }
            if not path_name in self.value['paths']:
                self.value['paths'][path_name] = {}
            self.value['paths'][path_name][verb.lower()] = path_value

    def set_title(self, title):
        '''Sets the name of the the API to be title'''
        basename = path.basename(title)
        self.value['info']['title'] = basename.split('.')[0] if basename.find('.') > 0 else basename

    def build(self):
        '''Returns the dict object representing the swagger specification.'''
        return self.value



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import json
import re
from tornado.httputil import parse_body_arguments

_named_param_regex = re.compile('(:([^/\s]*))')
FORM_URLENCODED = 'application/x-www-form-urlencoded'
MULTIPART_FORM_DATA = 'multipart/form-data'
APPLICATION_JSON = 'application/json'
TEXT_PLAIN = 'text/plain'

def format_request(expression):
    expression = json.dumps(expression)
    statement = "REQUEST = {}".format(expression)
    return statement

def parameterize_path(path):
    matches = re.findall(_named_param_regex, path)
    for match in matches:
        path = path.replace(match[0], '(?P<{}>[^\/]+)'.format(match[1]))
    return path.strip()

def parse_body(request):
    '''Takes an HTTP request and will parse the body depending on the Content-Type
    header. If no Content-Type is found, will treat the value as plain text. The
    return value is a dict or string representing the body.
    '''
    content_type = TEXT_PLAIN
    body = request.body
    body = '' if body is b'' or body is None else body.decode(encoding='UTF-8')
    if 'Content-Type' in request.headers:
        content_type = request.headers['Content-Type']
    return_body = body
    if content_type == FORM_URLENCODED or content_type.startswith(MULTIPART_FORM_DATA):
        # If there is form data, we already have the values in body_arguments, we
        # just need to convert the byte arrays to strings
        return_body = parse_args(request.body_arguments)
    elif content_type == APPLICATION_JSON:
        # Trying parsing the json, if we can't parse it the initial assignment
        # will treat the body as text
        try:
            return_body = json.loads(body)
        except Exception:
            pass
    return return_body

def parse_args(args):
    '''Converts args into a proper JSON string. args is expected to be a dictionary
    where the values are arrays of UTF-8 byte strings.
    '''
    ARGS = {}
    for key in args:
        ARGS[key] = []
        for value in args[key]:
            ARGS[key].append(value.decode(encoding='UTF-8'))
    return ARGS

def headers_to_dict(headers):
    new_headers = {}
    for header, header_value in headers.get_all():
        if header in new_headers:
            if not isinstance(new_headers[header], list):
                oringal_value = new_headers[header]
                new_headers[header] = [oringal_value]
            new_headers[header].append(header_value)
        else:
            new_headers[header] = header_value

    return new_headers



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
import re
import sys

def first_path_param_index(endpoint):
    '''Returns the index to the first path parameter for the endpoint. The
    index is not the string index, but rather where it is within the path.
    For example:
        first_path_param_index('/foo/:bar') # returns 1
        first_path_param_index('/foo/quo/:bar') # return 2
        first_path_param_index('/foo/quo/bar') # return sys.maxsize
    '''
    index = sys.maxsize
    if endpoint.find(':') >= 0:
        index = endpoint.count('/', 0, endpoint.find(':')) - 1
    return index

class APICellParser(object):
    '''
    A utility class for parsing Jupyter code cells in regards to using notebook
    cells as REST API endpoints.
    '''
    kernelspec_comment_mapping = {
        None:'#',
        'scala':'//'
    }
    api_indicator = '{}\s+(GET|PUT|POST|DELETE)\s+(\/.*)+'
    api_response_indicator = '{}\s+ResponseInfo\s+(GET|PUT|POST|DELETE)\s+(\/.*)+'

    def __init__(self, kernelspec):
        try:
            prefix = self.kernelspec_comment_mapping[kernelspec]
        except KeyError:
            prefix = self.kernelspec_comment_mapping[None]
        self.kernelspec_api_indicator = re.compile(self.api_indicator.format(prefix))
        self.kernelspec_api_response_indicator = re.compile(self.api_response_indicator.format(prefix))

    def is_api_cell(self, cell_source):
        '''Determines if the cell source is decroated to indicate an api cell'''
        match = self.kernelspec_api_indicator.match(cell_source)
        return match is not None

    def is_api_response_cell(self, cell_source):
        '''Determines if the cell source is decorated to indicate an api response cell'''
        match = self.kernelspec_api_response_indicator.match(cell_source)
        return match is not None

    def get_cell_endpoint_and_verb(self, cell_source):
        '''Parses a cell's source code and will return the endpoint and verb as a tuple'''
        endpoint = None
        verb = None
        matched = self.kernelspec_api_indicator.match(cell_source)
        if matched:
            endpoint = matched.group(2).strip()
            verb = matched.group(1)
        return endpoint, verb

    def endpoints(self, source_cells, sort_func=first_path_param_index):
        '''Return a list of tuples containing the method+URI and the cell source'''
        endpoints = {}
        for cell_source in source_cells:
            if self.is_api_cell(cell_source):
                matched = self.kernelspec_api_indicator.match(cell_source)
                uri = matched.group(2).strip()
                verb = matched.group(1)
                if uri not in endpoints:
                    endpoints[uri] = {}
                if verb not in endpoints[uri]:
                    endpoints[uri][verb] = cell_source
        sorted_keys = sorted(endpoints, key=sort_func, reverse=True)
        ret_val = []
        for key in sorted_keys:
            ret_val.append((key,endpoints[key]))
        return ret_val

    def endpoint_responses(self, source_cells, sort_func=first_path_param_index):
        '''Return a list of tuples containing the method+URI and the cell source'''
        endpoints = {}
        for cell_source in source_cells:
            if self.is_api_response_cell(cell_source):
                matched = self.kernelspec_api_response_indicator.match(cell_source)
                uri = matched.group(2).strip()
                verb = matched.group(1)
                if uri not in endpoints:
                    endpoints[uri] = {}
                if verb not in endpoints[uri]:
                    endpoints[uri][verb] = cell_source
        return endpoints



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import tornado
import notebook.services.sessions.handlers as notebook_handlers
from ...mixins import TokenAuthorizationMixin, CORSMixin, JSONErrorsMixin

class SessionRootHandler(TokenAuthorizationMixin,
                        CORSMixin,
                        JSONErrorsMixin,
                        notebook_handlers.SessionRootHandler):
    def get(self):
        '''
        Denies returning a list of running sessions  unless explicitly
        enabled, instead returning a 403 error indicating that the list is
        permanently forbidden.
        '''
        if 'kg_list_kernels' not in self.settings or self.settings['kg_list_kernels'] != True:
            raise tornado.web.HTTPError(403, 'Forbidden')
        else:
            super(SessionRootHandler, self).get()

default_handlers = []
for path, cls in notebook_handlers.default_handlers:
    if cls.__name__ in globals():
        # Use the same named class from here if it exists
        default_handlers.append((path, globals()[cls.__name__]))
    else:
        # Everything should have CORS and token auth
        bases = (TokenAuthorizationMixin, CORSMixin, cls)
        default_handlers.append((path, type(cls.__name__, bases, {})))



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
import tornado.web
import json
from tornado.log import access_log
from .request_utils import (parse_body, parse_args, format_request,
    headers_to_dict, parameterize_path)

from tornado import gen
from tornado.concurrent import Future
from ...mixins import TokenAuthorizationMixin, CORSMixin, JSONErrorsMixin
from functools import partial
from .errors import UnsupportedMethodError, CodeExecutionError
import os

class NotebookAPIHandler(TokenAuthorizationMixin,
                         CORSMixin,
                         JSONErrorsMixin,
                         tornado.web.RequestHandler):
    '''Executes annotated notebook cells in response to associated HTTP requests.'''

    def initialize(self, sources, response_sources, kernel_pool, kernel_name):
        self.kernel_pool = kernel_pool
        self.sources = sources
        self.kernel_name = kernel_name
        self.response_sources = response_sources

    def finish_future(self, future, result_accumulator):
        if result_accumulator['error']:
            future.set_exception(CodeExecutionError(result_accumulator['error']))
        elif len(result_accumulator['stream']) > 0:
            future.set_result(''.join(result_accumulator['stream']))
        elif result_accumulator['result']:
            future.set_result(json.dumps(result_accumulator['result']))
        else:
            # If nothing was set, return an empty value
            future.set_result('')

    def on_recv(self, result_accumulator, future, parent_header, msg):
        '''
        Receives messages for a particular code execution defined by parent_header.
        Collects all outputs from the kernel until an execution state of idle is received.
        This function should be used to generate partial functions where the newly derived
        function only takes msg as an argument.

        :param result_accumulator: Accumulates results across all messages received
        :param future: A future used to set the result (errors via exceptions)
        :param parent_header: The parent header value in the messages, indicating how
            results map to requests
        :param msg: The execution content/state message received.
        '''
        if msg['parent_header']['msg_id'] == parent_header:
            # On idle status, exit our loop
            if msg['header']['msg_type'] == 'status' and msg['content']['execution_state'] == 'idle':
                self.finish_future(future, result_accumulator)
            # Store the execute result
            elif msg['header']['msg_type'] == 'execute_result':
                result_accumulator['result'] = msg['content']['data']
            # Accumulate the stream messages
            elif msg['header']['msg_type'] == 'stream':
                # Only take stream output if it is on stdout or if the kernel
                # is non-confirming and does not name the stream
                if 'name' not in msg['content'] or msg['content']['name'] == 'stdout':
                    result_accumulator['stream'].append((msg['content']['text']))
            # Store the error message
            elif msg['header']['msg_type'] == 'error':
                error_name = msg['content']['ename']
                error_value = msg['content']['evalue']
                result_accumulator['error'] = 'Error {}: {} \n'.format(error_name, error_value)


    def execute_code(self, kernel_client, kernel_id, source_code):
        '''Executes source_code on the kernel specified and will return a Future to indicate when code
        execution is completed. If the code execution results in an error, a CodeExecutionError is raised.
        '''
        future = Future()
        result_accumulator = {'stream' : [], 'error' : None, 'result' : None}
        parent_header = kernel_client.execute(source_code)
        on_recv_func = partial(self.on_recv, result_accumulator, future, parent_header)
        self.kernel_pool.on_recv(kernel_id, on_recv_func)
        return future

    @gen.coroutine
    def _handle_request(self):
        '''Translates a HTTP request into code to execute on a kernel.'''
        self.response_future = Future()
        kernel_client, kernel_id = yield self.kernel_pool.acquire()
        try:
            # Method not supported
            if self.request.method not in self.sources:
                raise UnsupportedMethodError(self.request.method)

            # Set the Content-Type and status to default values
            self.set_header('Content-Type', 'text/plain')
            self.set_status(200)

            # Get the source to execute in response to this request
            source_code = self.sources[self.request.method]
            # Build the request dictionary
            request = json.dumps({
                'body' : parse_body(self.request),
                'args' : parse_args(self.request.query_arguments),
                'path' : self.path_kwargs,
                'headers' : headers_to_dict(self.request.headers)
            })
            # Turn the request string into a valid code string
            request_code = format_request(request)

            # Run the request and source code and yield until there's a result
            access_log.debug('Request code for notebook cell is: {}'.format(request_code))
            request_future = self.execute_code(kernel_client, kernel_id, request_code)
            yield request_future
            source_future = self.execute_code(kernel_client, kernel_id, source_code)
            source_result = yield source_future

            # If a response code cell exists, execute it
            if self.request.method in self.response_sources:
                response_code = self.response_sources[self.request.method]
                response_future = self.execute_code(kernel_client, kernel_id, response_code)

                # Wait for the response and parse the json value
                response_result = yield response_future
                response = json.loads(response_result)

                # Copy all the header values into the tornado response
                if 'headers' in response:
                    for header in response['headers']:
                        self.set_header(header, response['headers'][header])

                # Set the status code if it exists
                if 'status' in response:
                    self.set_status(response['status'])

            # Write the result of the source code execution
            self.write(source_result)
        # If there was a problem executing an code, return a 500
        except CodeExecutionError as err:
            self.write(err.error_message)
            self.set_status(500)
        # An unspported method was called on this handler
        except UnsupportedMethodError:
            self.set_status(405)
        finally:
            # Always make sure we release the kernel and finish the request
            self.response_future.set_result(None)
            self.kernel_pool.release(kernel_id)
            self.finish()

    @gen.coroutine
    def get(self, **kwargs):
        self._handle_request()
        yield self.response_future

    @gen.coroutine
    def post(self, **kwargs):
        self._handle_request()
        yield self.response_future

    @gen.coroutine
    def put(self, **kwargs):
        self._handle_request()
        yield self.response_future

    @gen.coroutine
    def delete(self, **kwargs):
        self._handle_request()
        yield self.response_future

    def options(self, **kwargs):
        self.finish()

class NotebookDownloadHandler(TokenAuthorizationMixin,
                              CORSMixin,
                              JSONErrorsMixin,
                              tornado.web.StaticFileHandler):
    '''Allows clients to download the original notebook behind the HTTP facade.'''
    def initialize(self, path):
        self.dirname, self.filename = os.path.split(path)
        super(NotebookDownloadHandler, self).initialize(self.dirname)

    def get(self, include_body=True):
        super(NotebookDownloadHandler, self).get(self.filename, include_body)



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
class CodeExecutionError(Exception):
    def __init__(self, error_message):
        self.error_message = error_message

class UnsupportedMethodError(Exception):
    def __init__(self, method):
        self.method = method



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import tornado.web
import json
from .manager import activity
from ...mixins import TokenAuthorizationMixin, CORSMixin, JSONErrorsMixin

class ActivityHandler(TokenAuthorizationMixin,
                      CORSMixin,
                      JSONErrorsMixin,
                      tornado.web.RequestHandler):

    def get(self, **kwargs):
        if not self.settings.get('kg_list_kernels'):
            raise tornado.web.HTTPError(403, 'Forbidden')
        self.set_header('Content-Type', 'application/json')
        self.set_status(200)
        self.finish(json.dumps(activity.get()))
